# Copyright (C) 2020 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unittests for the git_trace2_event_log.py module."""

import json
import os
import socket
import tempfile
import threading
import unittest
from unittest import mock

import git_trace2_event_log
import platform_utils


def serverLoggingThread(socket_path, server_ready, received_traces):
    """Helper function to receive logs over a Unix domain socket.

    Appends received messages on the provided socket and appends to
    received_traces.

    Args:
        socket_path: path to a Unix domain socket on which to listen for traces
        server_ready: a threading.Condition used to signal to the caller that
            this thread is ready to accept connections
        received_traces: a list to which received traces will be appended (after
            decoding to a utf-8 string).
    """
    platform_utils.remove(socket_path, missing_ok=True)
    data = b""
    with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as sock:
        sock.bind(socket_path)
        sock.listen(0)
        with server_ready:
            server_ready.notify()
        with sock.accept()[0] as conn:
            while True:
                recved = conn.recv(4096)
                if not recved:
                    break
                data += recved
    received_traces.extend(data.decode("utf-8").splitlines())


class EventLogTestCase(unittest.TestCase):
    """TestCase for the EventLog module."""

    PARENT_SID_KEY = "GIT_TRACE2_PARENT_SID"
    PARENT_SID_VALUE = "parent_sid"
    SELF_SID_REGEX = r"repo-\d+T\d+Z-.*"
    FULL_SID_REGEX = rf"^{PARENT_SID_VALUE}/{SELF_SID_REGEX}"

    def setUp(self):
        """Load the event_log module every time."""
        self._event_log_module = None
        # By default we initialize with the expected case where
        # repo launches us (so GIT_TRACE2_PARENT_SID is set).
        env = {
            self.PARENT_SID_KEY: self.PARENT_SID_VALUE,
        }
        self._event_log_module = git_trace2_event_log.EventLog(env=env)
        self._log_data = None

    def verifyCommonKeys(
        self, log_entry, expected_event_name=None, full_sid=True
    ):
        """Helper function to verify common event log keys."""
        self.assertIn("event", log_entry)
        self.assertIn("sid", log_entry)
        self.assertIn("thread", log_entry)
        self.assertIn("time", log_entry)

        # Do basic data format validation.
        if expected_event_name:
            self.assertEqual(expected_event_name, log_entry["event"])
        if full_sid:
            self.assertRegex(log_entry["sid"], self.FULL_SID_REGEX)
        else:
            self.assertRegex(log_entry["sid"], self.SELF_SID_REGEX)
        self.assertRegex(
            log_entry["time"], r"^\d+-\d+-\d+T\d+:\d+:\d+\.\d+\+00:00$"
        )

    def readLog(self, log_path):
        """Helper function to read log data into a list."""
        log_data = []
        with open(log_path, mode="rb") as f:
            for line in f:
                log_data.append(json.loads(line))
        return log_data

    def remove_prefix(self, s, prefix):
        """Return a copy string after removing |prefix| from |s|, if present or
        the original string."""
        if s.startswith(prefix):
            return s[len(prefix) :]
        else:
            return s

    def test_initial_state_with_parent_sid(self):
        """Test initial state when 'GIT_TRACE2_PARENT_SID' is set by parent."""
        self.assertRegex(self._event_log_module.full_sid, self.FULL_SID_REGEX)

    def test_initial_state_no_parent_sid(self):
        """Test initial state when 'GIT_TRACE2_PARENT_SID' is not set."""
        # Setup an empty environment dict (no parent sid).
        self._event_log_module = git_trace2_event_log.EventLog(env={})
        self.assertRegex(self._event_log_module.full_sid, self.SELF_SID_REGEX)

    def test_version_event(self):
        """Test 'version' event data is valid.

        Verify that the 'version' event is written even when no other
        events are addded.

        Expected event log:
        <version event>
        """
        with tempfile.TemporaryDirectory(prefix="event_log_tests") as tempdir:
            log_path = self._event_log_module.Write(path=tempdir)
            self._log_data = self.readLog(log_path)

        # A log with no added events should only have the version entry.
        self.assertEqual(len(self._log_data), 1)
        version_event = self._log_data[0]
        self.verifyCommonKeys(version_event, expected_event_name="version")
        # Check for 'version' event specific fields.
        self.assertIn("evt", version_event)
        self.assertIn("exe", version_event)
        # Verify "evt" version field is a string.
        self.assertIsInstance(version_event["evt"], str)

    def test_start_event(self):
        """Test and validate 'start' event data is valid.

        Expected event log:
        <version event>
        <start event>
        """
        self._event_log_module.StartEvent([])
        with tempfile.TemporaryDirectory(prefix="event_log_tests") as tempdir:
            log_path = self._event_log_module.Write(path=tempdir)
            self._log_data = self.readLog(log_path)

        self.assertEqual(len(self._log_data), 2)
        start_event = self._log_data[1]
        self.verifyCommonKeys(self._log_data[0], expected_event_name="version")
        self.verifyCommonKeys(start_event, expected_event_name="start")
        # Check for 'start' event specific fields.
        self.assertIn("argv", start_event)
        self.assertTrue(isinstance(start_event["argv"], list))

    def test_exit_event_result_none(self):
        """Test 'exit' event data is valid when result is None.

        We expect None result to be converted to 0 in the exit event data.

        Expected event log:
        <version event>
        <exit event>
        """
        self._event_log_module.ExitEvent(None)
        with tempfile.TemporaryDirectory(prefix="event_log_tests") as tempdir:
            log_path = self._event_log_module.Write(path=tempdir)
            self._log_data = self.readLog(log_path)

        self.assertEqual(len(self._log_data), 2)
        exit_event = self._log_data[1]
        self.verifyCommonKeys(self._log_data[0], expected_event_name="version")
        self.verifyCommonKeys(exit_event, expected_event_name="exit")
        # Check for 'exit' event specific fields.
        self.assertIn("code", exit_event)
        # 'None' result should convert to 0 (successful) return code.
        self.assertEqual(exit_event["code"], 0)

    def test_exit_event_result_integer(self):
        """Test 'exit' event data is valid when result is an integer.

        Expected event log:
        <version event>
        <exit event>
        """
        self._event_log_module.ExitEvent(2)
        with tempfile.TemporaryDirectory(prefix="event_log_tests") as tempdir:
            log_path = self._event_log_module.Write(path=tempdir)
            self._log_data = self.readLog(log_path)

        self.assertEqual(len(self._log_data), 2)
        exit_event = self._log_data[1]
        self.verifyCommonKeys(self._log_data[0], expected_event_name="version")
        self.verifyCommonKeys(exit_event, expected_event_name="exit")
        # Check for 'exit' event specific fields.
        self.assertIn("code", exit_event)
        self.assertEqual(exit_event["code"], 2)

    def test_command_event(self):
        """Test and validate 'command' event data is valid.

        Expected event log:
        <version event>
        <command event>
        """
        self._event_log_module.CommandEvent(
            name="repo", subcommands=["init", "this"]
        )
        with tempfile.TemporaryDirectory(prefix="event_log_tests") as tempdir:
            log_path = self._event_log_module.Write(path=tempdir)
            self._log_data = self.readLog(log_path)

        self.assertEqual(len(self._log_data), 2)
        command_event = self._log_data[1]
        self.verifyCommonKeys(self._log_data[0], expected_event_name="version")
        self.verifyCommonKeys(command_event, expected_event_name="cmd_name")
        # Check for 'command' event specific fields.
        self.assertIn("name", command_event)
        self.assertEqual(command_event["name"], "repo-init-this")

    def test_def_params_event_repo_config(self):
        """Test 'def_params' event data outputs only repo config keys.

        Expected event log:
        <version event>
        <def_param event>
        <def_param event>
        """
        config = {
            "git.foo": "bar",
            "repo.partialclone": "true",
            "repo.partialclonefilter": "blob:none",
        }
        self._event_log_module.DefParamRepoEvents(config)

        with tempfile.TemporaryDirectory(prefix="event_log_tests") as tempdir:
            log_path = self._event_log_module.Write(path=tempdir)
            self._log_data = self.readLog(log_path)

        self.assertEqual(len(self._log_data), 3)
        def_param_events = self._log_data[1:]
        self.verifyCommonKeys(self._log_data[0], expected_event_name="version")

        for event in def_param_events:
            self.verifyCommonKeys(event, expected_event_name="def_param")
            # Check for 'def_param' event specific fields.
            self.assertIn("param", event)
            self.assertIn("value", event)
            self.assertTrue(event["param"].startswith("repo."))

    def test_def_params_event_no_repo_config(self):
        """Test 'def_params' event data won't output non-repo config keys.

        Expected event log:
        <version event>
        """
        config = {
            "git.foo": "bar",
            "git.core.foo2": "baz",
        }
        self._event_log_module.DefParamRepoEvents(config)

        with tempfile.TemporaryDirectory(prefix="event_log_tests") as tempdir:
            log_path = self._event_log_module.Write(path=tempdir)
            self._log_data = self.readLog(log_path)

        self.assertEqual(len(self._log_data), 1)
        self.verifyCommonKeys(self._log_data[0], expected_event_name="version")

    def test_data_event_config(self):
        """Test 'data' event data outputs all config keys.

        Expected event log:
        <version event>
        <data event>
        <data event>
        """
        config = {
            "git.foo": "bar",
            "repo.partialclone": "false",
            "repo.syncstate.superproject.hassuperprojecttag": "true",
            "repo.syncstate.superproject.sys.argv": ["--", "sync", "protobuf"],
        }
        prefix_value = "prefix"
        self._event_log_module.LogDataConfigEvents(config, prefix_value)

        with tempfile.TemporaryDirectory(prefix="event_log_tests") as tempdir:
            log_path = self._event_log_module.Write(path=tempdir)
            self._log_data = self.readLog(log_path)

        self.assertEqual(len(self._log_data), 5)
        data_events = self._log_data[1:]
        self.verifyCommonKeys(self._log_data[0], expected_event_name="version")

        for event in data_events:
            self.verifyCommonKeys(event)
            # Check for 'data' event specific fields.
            self.assertIn("key", event)
            self.assertIn("value", event)
            key = event["key"]
            key = self.remove_prefix(key, f"{prefix_value}/")
            value = event["value"]
            self.assertEqual(
                self._event_log_module.GetDataEventName(value), event["event"]
            )
            self.assertTrue(key in config and value == config[key])

    def test_error_event(self):
        """Test and validate 'error' event data is valid.

        Expected event log:
        <version event>
        <error event>
        """
        msg = "invalid option: --cahced"
        fmt = "invalid option: %s"
        self._event_log_module.ErrorEvent(msg, fmt)
        with tempfile.TemporaryDirectory(prefix="event_log_tests") as tempdir:
            log_path = self._event_log_module.Write(path=tempdir)
            self._log_data = self.readLog(log_path)

        self.assertEqual(len(self._log_data), 2)
        error_event = self._log_data[1]
        self.verifyCommonKeys(self._log_data[0], expected_event_name="version")
        self.verifyCommonKeys(error_event, expected_event_name="error")
        # Check for 'error' event specific fields.
        self.assertIn("msg", error_event)
        self.assertIn("fmt", error_event)
        self.assertEqual(error_event["msg"], f"RepoErrorEvent:{msg}")
        self.assertEqual(error_event["fmt"], f"RepoErrorEvent:{fmt}")

    def test_write_with_filename(self):
        """Test Write() with a path to a file exits with None."""
        self.assertIsNone(self._event_log_module.Write(path="path/to/file"))

    def test_write_with_git_config(self):
        """Test Write() uses the git config path when 'git config' call
        succeeds."""
        with tempfile.TemporaryDirectory(prefix="event_log_tests") as tempdir:
            with mock.patch.object(
                self._event_log_module,
                "_GetEventTargetPath",
                return_value=tempdir,
            ):
                self.assertEqual(
                    os.path.dirname(self._event_log_module.Write()), tempdir
                )

    def test_write_no_git_config(self):
        """Test Write() with no git config variable present exits with None."""
        with mock.patch.object(
            self._event_log_module, "_GetEventTargetPath", return_value=None
        ):
            self.assertIsNone(self._event_log_module.Write())

    def test_write_non_string(self):
        """Test Write() with non-string type for |path| throws TypeError."""
        with self.assertRaises(TypeError):
            self._event_log_module.Write(path=1234)

    def test_write_socket(self):
        """Test Write() with Unix domain socket for |path| and validate received
        traces."""
        received_traces = []
        with tempfile.TemporaryDirectory(
            prefix="test_server_sockets"
        ) as tempdir:
            socket_path = os.path.join(tempdir, "server.sock")
            server_ready = threading.Condition()
            # Start "server" listening on Unix domain socket at socket_path.
            server_thread = threading.Thread(
                target=serverLoggingThread,
                args=(socket_path, server_ready, received_traces),
            )
            try:
                server_thread.start()

                with server_ready:
                    server_ready.wait(timeout=120)

                self._event_log_module.StartEvent([])
                path = self._event_log_module.Write(
                    path=f"af_unix:{socket_path}"
                )
            finally:
                server_thread.join(timeout=5)

        self.assertEqual(path, f"af_unix:stream:{socket_path}")
        self.assertEqual(len(received_traces), 2)
        version_event = json.loads(received_traces[0])
        start_event = json.loads(received_traces[1])
        self.verifyCommonKeys(version_event, expected_event_name="version")
        self.verifyCommonKeys(start_event, expected_event_name="start")
        # Check for 'start' event specific fields.
        self.assertIn("argv", start_event)
        self.assertIsInstance(start_event["argv"], list)



# Copyright (C) 2009 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import platform
import sys

from command import Command
from command import MirrorSafeCommand
from git_command import git
from git_command import RepoSourceVersion
from git_command import user_agent
from git_refs import HEAD
from wrapper import Wrapper


class Version(Command, MirrorSafeCommand):
    wrapper_version = None
    wrapper_path = None

    COMMON = False
    helpSummary = "Display the version of repo"
    helpUsage = """
%prog
"""

    def Execute(self, opt, args):
        rp = self.manifest.repoProject
        rem = rp.GetRemote()
        branch = rp.GetBranch("default")

        # These might not be the same.  Report them both.
        src_ver = RepoSourceVersion()
        rp_ver = rp.bare_git.describe(HEAD)
        print(f"repo version {rp_ver}")
        print(f"       (from {rem.url})")
        print(f"       (tracking {branch.merge})")
        print(f"       ({rp.bare_git.log('-1', '--format=%cD', HEAD)})")

        if self.wrapper_path is not None:
            print(f"repo launcher version {self.wrapper_version}")
            print(f"       (from {self.wrapper_path})")

            if src_ver != rp_ver:
                print(f"       (currently at {src_ver})")

        print(f"repo User-Agent {user_agent.repo}")
        print(f"git {git.version_tuple().full}")
        print(f"git User-Agent {user_agent.git}")
        print(f"Python {sys.version}")
        uname = platform.uname()
        if sys.version_info.major < 3:
            # Python 3 returns a named tuple, but Python 2 is simpler.
            print(uname)
        else:
            print(f"OS {uname.system} {uname.release} ({uname.version})")
            processor = uname.processor if uname.processor else "unknown"
            print(f"CPU {uname.machine} ({processor})")
        print("Bug reports:", Wrapper().BUG_URL)



# Copyright (C) 2010 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import sys

from command import Command
from error import GitError
from git_command import GitCommand
from repo_logging import RepoLogger


CHANGE_ID_RE = re.compile(r"^\s*Change-Id: I([0-9a-f]{40})\s*$")
logger = RepoLogger(__file__)


class CherryPick(Command):
    COMMON = True
    helpSummary = "Cherry-pick a change."
    helpUsage = """
%prog <sha1>
"""
    helpDescription = """
'%prog' cherry-picks a change from one branch to another.
The change id will be updated, and a reference to the old
change id will be added.
"""

    def ValidateOptions(self, opt, args):
        if len(args) != 1:
            self.Usage()

    def Execute(self, opt, args):
        reference = args[0]

        p = GitCommand(
            None,
            ["rev-parse", "--verify", reference],
            capture_stdout=True,
            capture_stderr=True,
            verify_command=True,
        )
        try:
            p.Wait()
        except GitError:
            logger.error(p.stderr)
            raise

        sha1 = p.stdout.strip()

        p = GitCommand(
            None,
            ["cat-file", "commit", sha1],
            capture_stdout=True,
            verify_command=True,
        )

        try:
            p.Wait()
        except GitError:
            logger.error("error: Failed to retrieve old commit message")
            raise

        old_msg = self._StripHeader(p.stdout)

        p = GitCommand(
            None,
            ["cherry-pick", sha1],
            capture_stdout=True,
            capture_stderr=True,
            verify_command=True,
        )

        try:
            p.Wait()
        except GitError as e:
            logger.error(e)
            logger.warning(
                "NOTE: When committing (please see above) and editing the "
                "commit message, please remove the old Change-Id-line and "
                "add:\n%s",
                self._GetReference(sha1),
            )
            raise

        if p.stdout:
            print(p.stdout.strip(), file=sys.stdout)
        if p.stderr:
            print(p.stderr.strip(), file=sys.stderr)

        # The cherry-pick was applied correctly. We just need to edit
        # the commit message.
        new_msg = self._Reformat(old_msg, sha1)

        p = GitCommand(
            None,
            ["commit", "--amend", "-F", "-"],
            input=new_msg,
            capture_stdout=True,
            capture_stderr=True,
            verify_command=True,
        )
        try:
            p.Wait()
        except GitError:
            logger.error("error: Failed to update commit message")
            raise

    def _IsChangeId(self, line):
        return CHANGE_ID_RE.match(line)

    def _GetReference(self, sha1):
        return "(cherry picked from commit %s)" % sha1

    def _StripHeader(self, commit_msg):
        lines = commit_msg.splitlines()
        return "\n".join(lines[lines.index("") + 1 :])

    def _Reformat(self, old_msg, sha1):
        new_msg = []

        for line in old_msg.splitlines():
            if not self._IsChangeId(line):
                new_msg.append(line)

        # Add a blank line between the message and the change id/reference.
        try:
            if new_msg[-1].strip() != "":
                new_msg.append("")
        except IndexError:
            pass

        new_msg.append(self._GetReference(sha1))
        return "\n".join(new_msg)



# Copyright (C) 2008 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import json
import os
import re
import subprocess
import sys
from typing import Any, Optional

from error import GitError
from error import RepoExitError
from git_refs import HEAD
from git_trace2_event_log_base import BaseEventLog
import platform_utils
from repo_logging import RepoLogger
from repo_trace import IsTrace
from repo_trace import REPO_TRACE
from repo_trace import Trace
from wrapper import Wrapper


GIT = "git"
GIT_DIR = "GIT_DIR"

LAST_GITDIR = None
LAST_CWD = None
DEFAULT_GIT_FAIL_MESSAGE = "git command failure"
ERROR_EVENT_LOGGING_PREFIX = "RepoGitCommandError"
# Common line length limit
GIT_ERROR_STDOUT_LINES = 1
GIT_ERROR_STDERR_LINES = 10
INVALID_GIT_EXIT_CODE = 126

logger = RepoLogger(__file__)


class _GitCall:
    @functools.lru_cache(maxsize=None)
    def version_tuple(self):
        ret = Wrapper().ParseGitVersion()
        if ret is None:
            msg = "fatal: unable to detect git version"
            logger.error(msg)
            raise GitRequireError(msg)
        return ret

    def __getattr__(self, name):
        name = name.replace("_", "-")

        def fun(*cmdv):
            command = [name]
            command.extend(cmdv)
            return GitCommand(None, command, add_event_log=False).Wait() == 0

        return fun


git = _GitCall()


def RepoSourceVersion():
    """Return the version of the repo.git tree."""
    ver = getattr(RepoSourceVersion, "version", None)

    # We avoid GitCommand so we don't run into circular deps -- GitCommand needs
    # to initialize version info we provide.
    if ver is None:
        env = GitCommand._GetBasicEnv()

        proj = os.path.dirname(os.path.abspath(__file__))
        env[GIT_DIR] = os.path.join(proj, ".git")
        result = subprocess.run(
            [GIT, "describe", HEAD],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            encoding="utf-8",
            env=env,
            check=False,
        )
        if result.returncode == 0:
            ver = result.stdout.strip()
            if ver.startswith("v"):
                ver = ver[1:]
        else:
            ver = "unknown"
        setattr(RepoSourceVersion, "version", ver)

    return ver


@functools.lru_cache(maxsize=None)
def GetEventTargetPath():
    """Get the 'trace2.eventtarget' path from git configuration.

    Returns:
        path: git config's 'trace2.eventtarget' path if it exists, or None
    """
    path = None
    cmd = ["config", "--get", "trace2.eventtarget"]
    # TODO(https://crbug.com/gerrit/13706): Use GitConfig when it supports
    # system git config variables.
    p = GitCommand(
        None,
        cmd,
        capture_stdout=True,
        capture_stderr=True,
        bare=True,
        add_event_log=False,
    )
    retval = p.Wait()
    if retval == 0:
        # Strip trailing carriage-return in path.
        path = p.stdout.rstrip("\n")
        if path == "":
            return None
    elif retval != 1:
        # `git config --get` is documented to produce an exit status of `1`
        # if the requested variable is not present in the configuration.
        # Report any other return value as an error.
        logger.error(
            "repo: error: 'git config --get' call failed with return code: "
            "%r, stderr: %r",
            retval,
            p.stderr,
        )
    return path


class UserAgent:
    """Mange User-Agent settings when talking to external services

    We follow the style as documented here:
    https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/User-Agent
    """

    _os = None
    _repo_ua = None
    _git_ua = None

    @property
    def os(self):
        """The operating system name."""
        if self._os is None:
            os_name = sys.platform
            if os_name.lower().startswith("linux"):
                os_name = "Linux"
            elif os_name == "win32":
                os_name = "Win32"
            elif os_name == "cygwin":
                os_name = "Cygwin"
            elif os_name == "darwin":
                os_name = "Darwin"
            self._os = os_name

        return self._os

    @property
    def repo(self):
        """The UA when connecting directly from repo."""
        if self._repo_ua is None:
            py_version = sys.version_info
            self._repo_ua = "git-repo/%s (%s) git/%s Python/%d.%d.%d" % (
                RepoSourceVersion(),
                self.os,
                git.version_tuple().full,
                py_version.major,
                py_version.minor,
                py_version.micro,
            )

        return self._repo_ua

    @property
    def git(self):
        """The UA when running git."""
        if self._git_ua is None:
            self._git_ua = (
                f"git/{git.version_tuple().full} ({self.os}) "
                f"git-repo/{RepoSourceVersion()}"
            )
        return self._git_ua


user_agent = UserAgent()


def git_require(min_version, fail=False, msg=""):
    git_version = git.version_tuple()
    if min_version <= git_version:
        return True
    if fail:
        need = ".".join(map(str, min_version))
        if msg:
            msg = " for " + msg
        error_msg = f"fatal: git {need} or later required{msg}"
        logger.error(error_msg)
        raise GitRequireError(error_msg)
    return False


def _build_env(
    _kwargs_only=(),
    bare: Optional[bool] = False,
    disable_editor: Optional[bool] = False,
    ssh_proxy: Optional[Any] = None,
    gitdir: Optional[str] = None,
    objdir: Optional[str] = None,
):
    """Constucts an env dict for command execution."""

    assert _kwargs_only == (), "_build_env only accepts keyword arguments."

    env = GitCommand._GetBasicEnv()

    if disable_editor:
        env["GIT_EDITOR"] = ":"
    if ssh_proxy:
        env["REPO_SSH_SOCK"] = ssh_proxy.sock()
        env["GIT_SSH"] = ssh_proxy.proxy
        env["GIT_SSH_VARIANT"] = "ssh"
    if "http_proxy" in env and "darwin" == sys.platform:
        s = f"'http.proxy={env['http_proxy']}'"
        p = env.get("GIT_CONFIG_PARAMETERS")
        if p is not None:
            s = p + " " + s
        env["GIT_CONFIG_PARAMETERS"] = s
    if "GIT_ALLOW_PROTOCOL" not in env:
        env[
            "GIT_ALLOW_PROTOCOL"
        ] = "file:git:http:https:ssh:persistent-http:persistent-https:sso:rpc"
    env["GIT_HTTP_USER_AGENT"] = user_agent.git

    if objdir:
        # Set to the place we want to save the objects.
        env["GIT_OBJECT_DIRECTORY"] = objdir

        alt_objects = os.path.join(gitdir, "objects") if gitdir else None
        if alt_objects and os.path.realpath(alt_objects) != os.path.realpath(
            objdir
        ):
            # Allow git to search the original place in case of local or unique
            # refs that git will attempt to resolve even if we aren't fetching
            # them.
            env["GIT_ALTERNATE_OBJECT_DIRECTORIES"] = alt_objects
    if bare and gitdir is not None:
        env[GIT_DIR] = gitdir

    return env


class GitCommand:
    """Wrapper around a single git invocation."""

    def __init__(
        self,
        project,
        cmdv,
        bare=False,
        input=None,
        capture_stdout=False,
        capture_stderr=False,
        merge_output=False,
        disable_editor=False,
        ssh_proxy=None,
        cwd=None,
        gitdir=None,
        objdir=None,
        verify_command=False,
        add_event_log=True,
        log_as_error=True,
    ):
        if project:
            if not cwd:
                cwd = project.worktree
            if not gitdir:
                gitdir = project.gitdir

        self.project = project
        self.cmdv = cmdv
        self.verify_command = verify_command
        self.stdout, self.stderr = None, None

        # Git on Windows wants its paths only using / for reliability.
        if platform_utils.isWindows():
            if objdir:
                objdir = objdir.replace("\\", "/")
            if gitdir:
                gitdir = gitdir.replace("\\", "/")

        env = _build_env(
            disable_editor=disable_editor,
            ssh_proxy=ssh_proxy,
            objdir=objdir,
            gitdir=gitdir,
            bare=bare,
        )

        command = [GIT]
        if bare:
            cwd = None
        command_name = cmdv[0]
        command.append(command_name)

        if command_name in ("fetch", "clone"):
            env["GIT_TERMINAL_PROMPT"] = "0"
            # Need to use the --progress flag for fetch/clone so output will be
            # displayed as by default git only does progress output if stderr is
            # a TTY.
            if sys.stderr.isatty():
                if "--progress" not in cmdv and "--quiet" not in cmdv:
                    command.append("--progress")
        command.extend(cmdv[1:])

        event_log = (
            BaseEventLog(env=env, add_init_count=True)
            if add_event_log
            else None
        )

        try:
            self._RunCommand(
                command,
                env,
                capture_stdout=capture_stdout,
                capture_stderr=capture_stderr,
                merge_output=merge_output,
                ssh_proxy=ssh_proxy,
                cwd=cwd,
                input=input,
            )
            self.VerifyCommand()
        except GitCommandError as e:
            if event_log is not None:
                error_info = json.dumps(
                    {
                        "ErrorType": type(e).__name__,
                        "Project": e.project,
                        "CommandName": command_name,
                        "Message": str(e),
                        "ReturnCode": str(e.git_rc)
                        if e.git_rc is not None
                        else None,
                        "IsError": log_as_error,
                    }
                )
                event_log.ErrorEvent(
                    f"{ERROR_EVENT_LOGGING_PREFIX}:{error_info}"
                )
                event_log.Write(GetEventTargetPath())
            if isinstance(e, GitPopenCommandError):
                raise

    def _RunCommand(
        self,
        command,
        env,
        capture_stdout=False,
        capture_stderr=False,
        merge_output=False,
        ssh_proxy=None,
        cwd=None,
        input=None,
    ):
        # Set subprocess.PIPE for streams that need to be captured.
        stdin = subprocess.PIPE if input else None
        stdout = subprocess.PIPE if capture_stdout else None
        stderr = (
            subprocess.STDOUT
            if merge_output
            else (subprocess.PIPE if capture_stderr else None)
        )

        # tee_stderr acts like a tee command for stderr, in that, it captures
        # stderr from the subprocess and streams it back to sys.stderr, while
        # keeping a copy in-memory.
        # This allows us to store stderr logs from the subprocess into
        # GitCommandError.
        # Certain git operations, such as `git push`, writes diagnostic logs,
        # such as, progress bar for pushing, into stderr. To ensure we don't
        # break git's UX, we need to write to sys.stderr as we read from the
        # subprocess. Setting encoding or errors makes subprocess return
        # io.TextIOWrapper, which is line buffered. To avoid line-buffering
        # while tee-ing stderr, we unset these kwargs. See GitCommand._Tee
        # for tee-ing between the streams.
        # We tee stderr iff the caller doesn't want to capture any stream to
        # not disrupt the existing flow.
        # See go/tee-repo-stderr for more context.
        tee_stderr = False
        kwargs = {"encoding": "utf-8", "errors": "backslashreplace"}
        if not (stdin or stdout or stderr):
            tee_stderr = True
            # stderr will be written back to sys.stderr even though it is
            # piped here.
            stderr = subprocess.PIPE
            kwargs = {}

        dbg = ""
        if IsTrace():
            global LAST_CWD
            global LAST_GITDIR

            if cwd and LAST_CWD != cwd:
                if LAST_GITDIR or LAST_CWD:
                    dbg += "\n"
                dbg += ": cd %s\n" % cwd
                LAST_CWD = cwd

            if GIT_DIR in env and LAST_GITDIR != env[GIT_DIR]:
                if LAST_GITDIR or LAST_CWD:
                    dbg += "\n"
                dbg += ": export GIT_DIR=%s\n" % env[GIT_DIR]
                LAST_GITDIR = env[GIT_DIR]

            if "GIT_OBJECT_DIRECTORY" in env:
                dbg += (
                    ": export GIT_OBJECT_DIRECTORY=%s\n"
                    % env["GIT_OBJECT_DIRECTORY"]
                )
            if "GIT_ALTERNATE_OBJECT_DIRECTORIES" in env:
                dbg += ": export GIT_ALTERNATE_OBJECT_DIRECTORIES=%s\n" % (
                    env["GIT_ALTERNATE_OBJECT_DIRECTORIES"]
                )

            dbg += ": "
            dbg += " ".join(command)
            if stdin == subprocess.PIPE:
                dbg += " 0<|"
            if stdout == subprocess.PIPE:
                dbg += " 1>|"
            if stderr == subprocess.PIPE:
                dbg += " 2>|"
            elif stderr == subprocess.STDOUT:
                dbg += " 2>&1"

        with Trace(
            "git command %s %s with debug: %s", LAST_GITDIR, command, dbg
        ):
            try:
                p = subprocess.Popen(
                    command,
                    cwd=cwd,
                    env=env,
                    stdin=stdin,
                    stdout=stdout,
                    stderr=stderr,
                    **kwargs,
                )
            except Exception as e:
                raise GitPopenCommandError(
                    message=f"{command[1]}: {e}",
                    project=self.project.name if self.project else None,
                    command_args=self.cmdv,
                )

            if ssh_proxy:
                ssh_proxy.add_client(p)

            self.process = p

            try:
                if tee_stderr:
                    # tee_stderr streams stderr to sys.stderr while capturing
                    # a copy within self.stderr. tee_stderr is only enabled
                    # when the caller wants to pipe no stream.
                    self.stderr = self._Tee(p.stderr, sys.stderr)
                else:
                    self.stdout, self.stderr = p.communicate(input=input)
            finally:
                if ssh_proxy:
                    ssh_proxy.remove_client(p)
            self.rc = p.wait()

    @staticmethod
    def _Tee(in_stream, out_stream):
        """Writes text from in_stream to out_stream while recording in buffer.

        Args:
            in_stream: I/O stream to be read from.
            out_stream: I/O stream to write to.

        Returns:
            A str containing everything read from the in_stream.
        """
        buffer = ""
        read_size = 1024 if sys.version_info < (3, 7) else -1
        chunk = in_stream.read1(read_size)
        while chunk:
            # Convert to str.
            if not hasattr(chunk, "encode"):
                chunk = chunk.decode("utf-8", "backslashreplace")

            buffer += chunk
            out_stream.write(chunk)
            out_stream.flush()

            chunk = in_stream.read1(read_size)

        return buffer

    @staticmethod
    def _GetBasicEnv():
        """Return a basic env for running git under.

        This is guaranteed to be side-effect free.
        """
        env = os.environ.copy()
        for key in (
            REPO_TRACE,
            GIT_DIR,
            "GIT_ALTERNATE_OBJECT_DIRECTORIES",
            "GIT_OBJECT_DIRECTORY",
            "GIT_WORK_TREE",
            "GIT_GRAFT_FILE",
            "GIT_INDEX_FILE",
        ):
            env.pop(key, None)
        return env

    def VerifyCommand(self):
        if self.rc == 0:
            return None
        stdout = (
            "\n".join(self.stdout.split("\n")[:GIT_ERROR_STDOUT_LINES])
            if self.stdout
            else None
        )
        stderr = (
            "\n".join(self.stderr.split("\n")[:GIT_ERROR_STDERR_LINES])
            if self.stderr
            else None
        )
        project = self.project.name if self.project else None
        raise GitCommandError(
            project=project,
            command_args=self.cmdv,
            git_rc=self.rc,
            git_stdout=stdout,
            git_stderr=stderr,
        )

    def Wait(self):
        if self.verify_command:
            self.VerifyCommand()
        return self.rc


class GitRequireError(RepoExitError):
    """Error raised when git version is unavailable or invalid."""

    def __init__(self, message, exit_code: int = INVALID_GIT_EXIT_CODE):
        super().__init__(message, exit_code=exit_code)


class GitCommandError(GitError):
    """
    Error raised from a failed git command.
    Note that GitError can refer to any Git related error (e.g. branch not
    specified for project.py 'UploadForReview'), while GitCommandError is
    raised exclusively from non-zero exit codes returned from git commands.
    """

    # Tuples with error formats and suggestions for those errors.
    _ERROR_TO_SUGGESTION = [
        (
            re.compile("couldn't find remote ref .*"),
            "Check if the provided ref exists in the remote.",
        ),
        (
            re.compile("unable to access '.*': .*"),
            (
                "Please make sure you have the correct access rights and the "
                "repository exists."
            ),
        ),
        (
            re.compile("'.*' does not appear to be a git repository"),
            "Are you running this repo command outside of a repo workspace?",
        ),
        (
            re.compile("not a git repository"),
            "Are you running this repo command outside of a repo workspace?",
        ),
    ]

    def __init__(
        self,
        message: str = DEFAULT_GIT_FAIL_MESSAGE,
        git_rc: int = None,
        git_stdout: str = None,
        git_stderr: str = None,
        **kwargs,
    ):
        super().__init__(
            message,
            **kwargs,
        )
        self.git_rc = git_rc
        self.git_stdout = git_stdout
        self.git_stderr = git_stderr

    @property
    @functools.lru_cache(maxsize=None)
    def suggestion(self):
        """Returns helpful next steps for the given stderr."""
        if not self.git_stderr:
            return self.git_stderr

        for err, suggestion in self._ERROR_TO_SUGGESTION:
            if err.search(self.git_stderr):
                return suggestion

        return None

    def __str__(self):
        args = "[]" if not self.command_args else " ".join(self.command_args)
        error_type = type(self).__name__
        string = f"{error_type}: '{args}' on {self.project} failed"

        if self.message != DEFAULT_GIT_FAIL_MESSAGE:
            string += f": {self.message}"

        if self.git_stdout:
            string += f"\nstdout: {self.git_stdout}"

        if self.git_stderr:
            string += f"\nstderr: {self.git_stderr}"

        if self.suggestion:
            string += f"\nsuggestion: {self.suggestion}"

        return string


class GitPopenCommandError(GitError):
    """
    Error raised when subprocess.Popen fails for a GitCommand
    """



# Copyright (C) 2021 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains functions used to fetch files from various sources."""

import subprocess
import sys
from urllib.parse import urlparse
from urllib.request import urlopen

from error import RepoExitError


class FetchFileError(RepoExitError):
    """Exit error when fetch_file fails."""


def fetch_file(url, verbose=False):
    """Fetch a file from the specified source using the appropriate protocol.

    Returns:
        The contents of the file as bytes.
    """
    scheme = urlparse(url).scheme
    if scheme == "gs":
        cmd = ["gsutil", "cat", url]
        errors = []
        try:
            result = subprocess.run(
                cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True
            )
            if result.stderr and verbose:
                print(
                    'warning: non-fatal error running "gsutil": %s'
                    % result.stderr,
                    file=sys.stderr,
                )
            return result.stdout
        except subprocess.CalledProcessError as e:
            errors.append(e)
            print(
                'fatal: error running "gsutil": %s' % e.stderr, file=sys.stderr
            )
        raise FetchFileError(aggregate_errors=errors)
    with urlopen(url) as f:
        return f.read()



# Copyright (C) 2017 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import multiprocessing


TASK_COMMAND = "command"
TASK_SYNC_NETWORK = "sync-network"
TASK_SYNC_LOCAL = "sync-local"


class EventLog:
    """Event log that records events that occurred during a repo invocation.

    Events are written to the log as a consecutive JSON entries, one per line.
    Each entry contains the following keys:
    - id: A ('RepoOp', ID) tuple, suitable for storing in a datastore.
          The ID is only unique for the invocation of the repo command.
    - name: Name of the object being operated upon.
    - task_name: The task that was performed.
    - start: Timestamp of when the operation started.
    - finish: Timestamp of when the operation finished.
    - success: Boolean indicating if the operation was successful.
    - try_count: A counter indicating the try count of this task.

    Optionally:
    - parent: A ('RepoOp', ID) tuple indicating the parent event for nested
        events.

    Valid task_names include:
    - command: The invocation of a subcommand.
    - sync-network: The network component of a sync command.
    - sync-local: The local component of a sync command.

    Specific tasks may include additional informational properties.
    """

    def __init__(self):
        """Initializes the event log."""
        self._log = []
        self._parent = None

    def Add(
        self,
        name,
        task_name,
        start,
        finish=None,
        success=None,
        try_count=1,
        kind="RepoOp",
    ):
        """Add an event to the log.

        Args:
            name: Name of the object being operated upon.
            task_name: A sub-task that was performed for name.
            start: Timestamp of when the operation started.
            finish: Timestamp of when the operation finished.
            success: Boolean indicating if the operation was successful.
            try_count: A counter indicating the try count of this task.
            kind: The kind of the object for the unique identifier.

        Returns:
            A dictionary of the event added to the log.
        """
        event = {
            "id": (kind, _NextEventId()),
            "name": name,
            "task_name": task_name,
            "start_time": start,
            "try": try_count,
        }

        if self._parent:
            event["parent"] = self._parent["id"]

        if success is not None or finish is not None:
            self.FinishEvent(event, finish, success)

        self._log.append(event)
        return event

    def AddSync(self, project, task_name, start, finish, success):
        """Add a event to the log for a sync command.

        Args:
            project: Project being synced.
            task_name: A sub-task that was performed for name.
                One of (TASK_SYNC_NETWORK, TASK_SYNC_LOCAL)
            start: Timestamp of when the operation started.
            finish: Timestamp of when the operation finished.
            success: Boolean indicating if the operation was successful.

        Returns:
            A dictionary of the event added to the log.
        """
        event = self.Add(project.relpath, task_name, start, finish, success)
        if event is not None:
            event["project"] = project.name
            if project.revisionExpr:
                event["revision"] = project.revisionExpr
            if project.remote.url:
                event["project_url"] = project.remote.url
            if project.remote.fetchUrl:
                event["remote_url"] = project.remote.fetchUrl
            try:
                event["git_hash"] = project.GetCommitRevisionId()
            except Exception:
                pass
        return event

    def GetStatusString(self, success):
        """Converst a boolean success to a status string.

        Args:
            success: Boolean indicating if the operation was successful.

        Returns:
            status string.
        """
        return "pass" if success else "fail"

    def FinishEvent(self, event, finish, success):
        """Finishes an incomplete event.

        Args:
            event: An event that has been added to the log.
            finish: Timestamp of when the operation finished.
            success: Boolean indicating if the operation was successful.

        Returns:
            A dictionary of the event added to the log.
        """
        event["status"] = self.GetStatusString(success)
        event["finish_time"] = finish
        return event

    def SetParent(self, event):
        """Set a parent event for all new entities.

        Args:
            event: The event to use as a parent.
        """
        self._parent = event

    def Write(self, filename):
        """Writes the log out to a file.

        Args:
            filename: The file to write the log to.
        """
        with open(filename, "w+") as f:
            for e in self._log:
                json.dump(e, f, sort_keys=True)
                f.write("\n")


# An integer id that is unique across this invocation of the program, to be set
# by the first Add event. We can't set it here since it results in leaked
# resources (see: https://issues.gerritcodereview.com/353656374).
_EVENT_ID = None


def _NextEventId():
    """Helper function for grabbing the next unique id.

    Returns:
        A unique, to this invocation of the program, integer id.
    """
    global _EVENT_ID
    if _EVENT_ID is None:
        # There is a small chance of race condition - two parallel processes
        # setting up _EVENT_ID. However, we expect TASK_COMMAND to happen before
        # mp kicks in.
        _EVENT_ID = multiprocessing.Value("i", 1)
    with _EVENT_ID.get_lock():
        val = _EVENT_ID.value
        _EVENT_ID.value += 1
    return val



# Copyright (C) 2012 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import optparse

from color import Coloring
from command import PagedCommand


class Overview(PagedCommand):
    COMMON = True
    helpSummary = "Display overview of unmerged project branches"
    helpUsage = """
%prog [--current-branch] [<project>...]
"""
    helpDescription = """
The '%prog' command is used to display an overview of the projects branches,
and list any local commits that have not yet been merged into the project.

The -c/--current-branch option can be used to restrict the output to only
branches currently checked out in each project.  By default, all branches
are displayed.
"""

    def _Options(self, p):
        p.add_option(
            "-c",
            "--current-branch",
            dest="current_branch",
            action="store_true",
            help="consider only checked out branches",
        )
        p.add_option(
            "--no-current-branch",
            dest="current_branch",
            action="store_false",
            help="consider all local branches",
        )
        # Turn this into a warning & remove this someday.
        p.add_option(
            "-b",
            dest="current_branch",
            action="store_true",
            help=optparse.SUPPRESS_HELP,
        )

    def Execute(self, opt, args):
        all_branches = []
        for project in self.GetProjects(
            args, all_manifests=not opt.this_manifest_only
        ):
            br = [project.GetUploadableBranch(x) for x in project.GetBranches()]
            br = [x for x in br if x]
            if opt.current_branch:
                br = [x for x in br if x.name == project.CurrentBranch]
            all_branches.extend(br)

        if not all_branches:
            return

        class Report(Coloring):
            def __init__(self, config):
                Coloring.__init__(self, config, "status")
                self.project = self.printer("header", attr="bold")
                self.text = self.printer("text")

        out = Report(all_branches[0].project.config)
        out.text("Deprecated. See repo info -o.")
        out.nl()
        out.project("Projects Overview")
        out.nl()

        project = None

        for branch in all_branches:
            if project != branch.project:
                project = branch.project
                out.nl()
                out.project(
                    "project %s/"
                    % project.RelPath(local=opt.this_manifest_only)
                )
                out.nl()

            commits = branch.commits
            date = branch.date
            print(
                "%s %-33s (%2d commit%s, %s)"
                % (
                    branch.name == project.CurrentBranch and "*" or " ",
                    branch.name,
                    len(commits),
                    len(commits) != 1 and "s" or " ",
                    date,
                )
            )
            for commit in commits:
                print("%-35s   - %s" % ("", commit))



# Copyright (C) 2008 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List


class BaseRepoError(Exception):
    """All repo specific exceptions derive from BaseRepoError."""


class RepoError(BaseRepoError):
    """Exceptions thrown inside repo that can be handled."""

    def __init__(self, *args, project: str = None) -> None:
        super().__init__(*args)
        self.project = project


class RepoExitError(BaseRepoError):
    """Exception thrown that result in termination of repo program.
    - Should only be handled in main.py
    """

    def __init__(
        self,
        *args,
        exit_code: int = 1,
        aggregate_errors: List[Exception] = None,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.exit_code = exit_code
        self.aggregate_errors = aggregate_errors


class RepoUnhandledExceptionError(RepoExitError):
    """Exception that maintains error as reason for program exit."""

    def __init__(
        self,
        error: BaseException,
        **kwargs,
    ) -> None:
        super().__init__(error, **kwargs)
        self.error = error


class SilentRepoExitError(RepoExitError):
    """RepoExitError that should no include CLI logging of issue/issues."""


class ManifestParseError(RepoExitError):
    """Failed to parse the manifest file."""


class ManifestInvalidRevisionError(ManifestParseError):
    """The revision value in a project is incorrect."""


class ManifestInvalidPathError(ManifestParseError):
    """A path used in <copyfile> or <linkfile> is incorrect."""


class NoManifestException(RepoExitError):
    """The required manifest does not exist."""

    def __init__(self, path, reason, **kwargs):
        super().__init__(path, reason, **kwargs)
        self.path = path
        self.reason = reason

    def __str__(self):
        return self.reason


class EditorError(RepoError):
    """Unspecified error from the user's text editor."""

    def __init__(self, reason, **kwargs):
        super().__init__(reason, **kwargs)
        self.reason = reason

    def __str__(self):
        return self.reason


class GitError(RepoError):
    """Unspecified git related error."""

    def __init__(self, message, command_args=None, **kwargs):
        super().__init__(message, **kwargs)
        self.message = message
        self.command_args = command_args

    def __str__(self):
        return self.message


class GitAuthError(RepoExitError):
    """Cannot talk to remote due to auth issue."""


class UploadError(RepoError):
    """A bundle upload to Gerrit did not succeed."""

    def __init__(self, reason, **kwargs):
        super().__init__(reason, **kwargs)
        self.reason = reason

    def __str__(self):
        return self.reason


class DownloadError(RepoExitError):
    """Cannot download a repository."""

    def __init__(self, reason, **kwargs):
        super().__init__(reason, **kwargs)
        self.reason = reason

    def __str__(self):
        return self.reason


class InvalidArgumentsError(RepoExitError):
    """Invalid command Arguments."""


class SyncError(RepoExitError):
    """Cannot sync repo."""


class UpdateManifestError(RepoExitError):
    """Cannot update manifest."""


class NoSuchProjectError(RepoExitError):
    """A specified project does not exist in the work tree."""

    def __init__(self, name=None, **kwargs):
        super().__init__(**kwargs)
        self.name = name

    def __str__(self):
        if self.name is None:
            return "in current directory"
        return self.name


class InvalidProjectGroupsError(RepoExitError):
    """A specified project is not suitable for the specified groups"""

    def __init__(self, name=None, **kwargs):
        super().__init__(**kwargs)
        self.name = name

    def __str__(self):
        if self.name is None:
            return "in current directory"
        return self.name


class RepoChangedException(BaseRepoError):
    """Thrown if 'repo sync' results in repo updating its internal
    repo or manifest repositories.  In this special case we must
    use exec to re-execute repo with the new code and manifest.
    """

    def __init__(self, extra_args=None):
        super().__init__(extra_args)
        self.extra_args = extra_args or []


class HookError(RepoError):
    """Thrown if a 'repo-hook' could not be run.

    The common case is that the file wasn't present when we tried to run it.
    """



# Copyright (C) 2008 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os


# A mapping of the subcommand name to the class that implements it.
all_commands = {}
all_modules = []

my_dir = os.path.dirname(__file__)
for py in os.listdir(my_dir):
    if py == "__init__.py":
        continue

    if py.endswith(".py"):
        name = py[:-3]

        clsn = name.capitalize()
        while clsn.find("_") > 0:
            h = clsn.index("_")
            clsn = clsn[0:h] + clsn[h + 1 :].capitalize()

        mod = __import__(__name__, globals(), locals(), ["%s" % name])
        mod = getattr(mod, name)
        try:
            cmd = getattr(mod, clsn)
        except AttributeError:
            raise SyntaxError(f"{__name__}/{py} does not define class {clsn}")

        name = name.replace("_", "-")
        cmd.NAME = name
        all_commands[name] = cmd
        all_modules.append(mod)

# Add 'branch' as an alias for 'branches'.
all_commands["branch"] = all_commands["branches"]



# Copyright (C) 2020 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Random utility code for release tools."""

import os
import re
import subprocess
import sys


assert sys.version_info >= (3, 6), "This module requires Python 3.6+"


TOPDIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
HOMEDIR = os.path.expanduser("~")


# These are the release keys we sign with.
KEYID_DSA = "8BB9AD793E8E6153AF0F9A4416530D5E920F5C65"
KEYID_RSA = "A34A13BE8E76BFF46A0C022DA2E75A824AAB9624"
KEYID_ECC = "E1F9040D7A3F6DAFAC897CD3D3B95DA243E48A39"


def cmdstr(cmd):
    """Get a nicely quoted shell command."""
    ret = []
    for arg in cmd:
        if not re.match(r"^[a-zA-Z0-9/_.=-]+$", arg):
            arg = f'"{arg}"'
        ret.append(arg)
    return " ".join(ret)


def run(opts, cmd, check=True, **kwargs):
    """Helper around subprocess.run to include logging."""
    print("+", cmdstr(cmd))
    if opts.dryrun:
        cmd = ["true", "--"] + cmd
    try:
        return subprocess.run(cmd, check=check, **kwargs)
    except subprocess.CalledProcessError as e:
        print(f"aborting: {e}", file=sys.stderr)
        sys.exit(1)


def import_release_key(opts):
    """Import the public key of the official release repo signing key."""
    # Extract the key from our repo launcher.
    launcher = getattr(opts, "launcher", os.path.join(TOPDIR, "repo"))
    print(f'Importing keys from "{launcher}" launcher script')
    with open(launcher, encoding="utf-8") as fp:
        data = fp.read()

    keys = re.findall(
        r"\n-----BEGIN PGP PUBLIC KEY BLOCK-----\n[^-]*"
        r"\n-----END PGP PUBLIC KEY BLOCK-----\n",
        data,
        flags=re.M,
    )
    run(opts, ["gpg", "--import"], input="\n".join(keys).encode("utf-8"))

    print("Marking keys as fully trusted")
    run(
        opts,
        ["gpg", "--import-ownertrust"],
        input=f"{KEYID_DSA}:6:\n".encode("utf-8"),
    )



# Copyright (C) 2015 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unittests for the wrapper.py module."""

import io
import os
import re
import sys
import tempfile
import unittest
from unittest import mock

import git_command
import main
import wrapper


def fixture(*paths):
    """Return a path relative to tests/fixtures."""
    return os.path.join(os.path.dirname(__file__), "fixtures", *paths)


class RepoWrapperTestCase(unittest.TestCase):
    """TestCase for the wrapper module."""

    def setUp(self):
        """Load the wrapper module every time."""
        wrapper.Wrapper.cache_clear()
        self.wrapper = wrapper.Wrapper()


class RepoWrapperUnitTest(RepoWrapperTestCase):
    """Tests helper functions in the repo wrapper"""

    def test_version(self):
        """Make sure _Version works."""
        with self.assertRaises(SystemExit) as e:
            with mock.patch("sys.stdout", new_callable=io.StringIO) as stdout:
                with mock.patch(
                    "sys.stderr", new_callable=io.StringIO
                ) as stderr:
                    self.wrapper._Version()
        self.assertEqual(0, e.exception.code)
        self.assertEqual("", stderr.getvalue())
        self.assertIn("repo launcher version", stdout.getvalue())

    def test_python_constraints(self):
        """The launcher should never require newer than main.py."""
        self.assertGreaterEqual(
            main.MIN_PYTHON_VERSION_HARD, self.wrapper.MIN_PYTHON_VERSION_HARD
        )
        self.assertGreaterEqual(
            main.MIN_PYTHON_VERSION_SOFT, self.wrapper.MIN_PYTHON_VERSION_SOFT
        )
        # Make sure the versions are themselves in sync.
        self.assertGreaterEqual(
            self.wrapper.MIN_PYTHON_VERSION_SOFT,
            self.wrapper.MIN_PYTHON_VERSION_HARD,
        )

    def test_init_parser(self):
        """Make sure 'init' GetParser works."""
        parser = self.wrapper.GetParser()
        opts, args = parser.parse_args([])
        self.assertEqual([], args)
        self.assertIsNone(opts.manifest_url)


class SetGitTrace2ParentSid(RepoWrapperTestCase):
    """Check SetGitTrace2ParentSid behavior."""

    KEY = "GIT_TRACE2_PARENT_SID"
    VALID_FORMAT = re.compile(r"^repo-[0-9]{8}T[0-9]{6}Z-P[0-9a-f]{8}$")

    def test_first_set(self):
        """Test env var not yet set."""
        env = {}
        self.wrapper.SetGitTrace2ParentSid(env)
        self.assertIn(self.KEY, env)
        value = env[self.KEY]
        self.assertRegex(value, self.VALID_FORMAT)

    def test_append(self):
        """Test env var is appended."""
        env = {self.KEY: "pfx"}
        self.wrapper.SetGitTrace2ParentSid(env)
        self.assertIn(self.KEY, env)
        value = env[self.KEY]
        self.assertTrue(value.startswith("pfx/"))
        self.assertRegex(value[4:], self.VALID_FORMAT)

    def test_global_context(self):
        """Check os.environ gets updated by default."""
        os.environ.pop(self.KEY, None)
        self.wrapper.SetGitTrace2ParentSid()
        self.assertIn(self.KEY, os.environ)
        value = os.environ[self.KEY]
        self.assertRegex(value, self.VALID_FORMAT)


class RunCommand(RepoWrapperTestCase):
    """Check run_command behavior."""

    def test_capture(self):
        """Check capture_output handling."""
        ret = self.wrapper.run_command(["echo", "hi"], capture_output=True)
        # echo command appends OS specific linesep, but on Windows + Git Bash
        # we get UNIX ending, so we allow both.
        self.assertIn(ret.stdout, ["hi" + os.linesep, "hi\n"])

    def test_check(self):
        """Check check handling."""
        self.wrapper.run_command(["true"], check=False)
        self.wrapper.run_command(["true"], check=True)
        self.wrapper.run_command(["false"], check=False)
        with self.assertRaises(self.wrapper.RunError):
            self.wrapper.run_command(["false"], check=True)


class RunGit(RepoWrapperTestCase):
    """Check run_git behavior."""

    def test_capture(self):
        """Check capture_output handling."""
        ret = self.wrapper.run_git("--version")
        self.assertIn("git", ret.stdout)

    def test_check(self):
        """Check check handling."""
        with self.assertRaises(self.wrapper.CloneFailure):
            self.wrapper.run_git("--version-asdfasdf")
        self.wrapper.run_git("--version-asdfasdf", check=False)


class ParseGitVersion(RepoWrapperTestCase):
    """Check ParseGitVersion behavior."""

    def test_autoload(self):
        """Check we can load the version from the live git."""
        ret = self.wrapper.ParseGitVersion()
        self.assertIsNotNone(ret)

    def test_bad_ver(self):
        """Check handling of bad git versions."""
        ret = self.wrapper.ParseGitVersion(ver_str="asdf")
        self.assertIsNone(ret)

    def test_normal_ver(self):
        """Check handling of normal git versions."""
        ret = self.wrapper.ParseGitVersion(ver_str="git version 2.25.1")
        self.assertEqual(2, ret.major)
        self.assertEqual(25, ret.minor)
        self.assertEqual(1, ret.micro)
        self.assertEqual("2.25.1", ret.full)

    def test_extended_ver(self):
        """Check handling of extended distro git versions."""
        ret = self.wrapper.ParseGitVersion(
            ver_str="git version 1.30.50.696.g5e7596f4ac-goog"
        )
        self.assertEqual(1, ret.major)
        self.assertEqual(30, ret.minor)
        self.assertEqual(50, ret.micro)
        self.assertEqual("1.30.50.696.g5e7596f4ac-goog", ret.full)


class CheckGitVersion(RepoWrapperTestCase):
    """Check _CheckGitVersion behavior."""

    def test_unknown(self):
        """Unknown versions should abort."""
        with mock.patch.object(
            self.wrapper, "ParseGitVersion", return_value=None
        ):
            with self.assertRaises(self.wrapper.CloneFailure):
                self.wrapper._CheckGitVersion()

    def test_old(self):
        """Old versions should abort."""
        with mock.patch.object(
            self.wrapper,
            "ParseGitVersion",
            return_value=self.wrapper.GitVersion(1, 0, 0, "1.0.0"),
        ):
            with self.assertRaises(self.wrapper.CloneFailure):
                self.wrapper._CheckGitVersion()

    def test_new(self):
        """Newer versions should run fine."""
        with mock.patch.object(
            self.wrapper,
            "ParseGitVersion",
            return_value=self.wrapper.GitVersion(100, 0, 0, "100.0.0"),
        ):
            self.wrapper._CheckGitVersion()


class Requirements(RepoWrapperTestCase):
    """Check Requirements handling."""

    def test_missing_file(self):
        """Don't crash if the file is missing (old version)."""
        testdir = os.path.dirname(os.path.realpath(__file__))
        self.assertIsNone(self.wrapper.Requirements.from_dir(testdir))
        self.assertIsNone(
            self.wrapper.Requirements.from_file(
                os.path.join(testdir, "xxxxxxxxxxxxxxxxxxxxxxxx")
            )
        )

    def test_corrupt_data(self):
        """If the file can't be parsed, don't blow up."""
        self.assertIsNone(self.wrapper.Requirements.from_file(__file__))
        self.assertIsNone(self.wrapper.Requirements.from_data(b"x"))

    def test_valid_data(self):
        """Make sure we can parse the file we ship."""
        self.assertIsNotNone(self.wrapper.Requirements.from_data(b"{}"))
        rootdir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        self.assertIsNotNone(self.wrapper.Requirements.from_dir(rootdir))
        self.assertIsNotNone(
            self.wrapper.Requirements.from_file(
                os.path.join(rootdir, "requirements.json")
            )
        )

    def test_format_ver(self):
        """Check format_ver can format."""
        self.assertEqual(
            "1.2.3", self.wrapper.Requirements._format_ver((1, 2, 3))
        )
        self.assertEqual("1", self.wrapper.Requirements._format_ver([1]))

    def test_assert_all_unknown(self):
        """Check assert_all works with incompatible file."""
        reqs = self.wrapper.Requirements({})
        reqs.assert_all()

    def test_assert_all_new_repo(self):
        """Check assert_all accepts new enough repo."""
        reqs = self.wrapper.Requirements({"repo": {"hard": [1, 0]}})
        reqs.assert_all()

    def test_assert_all_old_repo(self):
        """Check assert_all rejects old repo."""
        reqs = self.wrapper.Requirements({"repo": {"hard": [99999, 0]}})
        with self.assertRaises(SystemExit):
            reqs.assert_all()

    def test_assert_all_new_python(self):
        """Check assert_all accepts new enough python."""
        reqs = self.wrapper.Requirements({"python": {"hard": sys.version_info}})
        reqs.assert_all()

    def test_assert_all_old_python(self):
        """Check assert_all rejects old python."""
        reqs = self.wrapper.Requirements({"python": {"hard": [99999, 0]}})
        with self.assertRaises(SystemExit):
            reqs.assert_all()

    def test_assert_ver_unknown(self):
        """Check assert_ver works with incompatible file."""
        reqs = self.wrapper.Requirements({})
        reqs.assert_ver("xxx", (1, 0))

    def test_assert_ver_new(self):
        """Check assert_ver allows new enough versions."""
        reqs = self.wrapper.Requirements(
            {"git": {"hard": [1, 0], "soft": [2, 0]}}
        )
        reqs.assert_ver("git", (1, 0))
        reqs.assert_ver("git", (1, 5))
        reqs.assert_ver("git", (2, 0))
        reqs.assert_ver("git", (2, 5))

    def test_assert_ver_old(self):
        """Check assert_ver rejects old versions."""
        reqs = self.wrapper.Requirements(
            {"git": {"hard": [1, 0], "soft": [2, 0]}}
        )
        with self.assertRaises(SystemExit):
            reqs.assert_ver("git", (0, 5))


class NeedSetupGnuPG(RepoWrapperTestCase):
    """Check NeedSetupGnuPG behavior."""

    def test_missing_dir(self):
        """The ~/.repoconfig tree doesn't exist yet."""
        with tempfile.TemporaryDirectory(prefix="repo-tests") as tempdir:
            self.wrapper.home_dot_repo = os.path.join(tempdir, "foo")
            self.assertTrue(self.wrapper.NeedSetupGnuPG())

    def test_missing_keyring(self):
        """The keyring-version file doesn't exist yet."""
        with tempfile.TemporaryDirectory(prefix="repo-tests") as tempdir:
            self.wrapper.home_dot_repo = tempdir
            self.assertTrue(self.wrapper.NeedSetupGnuPG())

    def test_empty_keyring(self):
        """The keyring-version file exists, but is empty."""
        with tempfile.TemporaryDirectory(prefix="repo-tests") as tempdir:
            self.wrapper.home_dot_repo = tempdir
            with open(os.path.join(tempdir, "keyring-version"), "w"):
                pass
            self.assertTrue(self.wrapper.NeedSetupGnuPG())

    def test_old_keyring(self):
        """The keyring-version file exists, but it's old."""
        with tempfile.TemporaryDirectory(prefix="repo-tests") as tempdir:
            self.wrapper.home_dot_repo = tempdir
            with open(os.path.join(tempdir, "keyring-version"), "w") as fp:
                fp.write("1.0\n")
            self.assertTrue(self.wrapper.NeedSetupGnuPG())

    def test_new_keyring(self):
        """The keyring-version file exists, and is up-to-date."""
        with tempfile.TemporaryDirectory(prefix="repo-tests") as tempdir:
            self.wrapper.home_dot_repo = tempdir
            with open(os.path.join(tempdir, "keyring-version"), "w") as fp:
                fp.write("1000.0\n")
            self.assertFalse(self.wrapper.NeedSetupGnuPG())


class SetupGnuPG(RepoWrapperTestCase):
    """Check SetupGnuPG behavior."""

    def test_full(self):
        """Make sure it works completely."""
        with tempfile.TemporaryDirectory(prefix="repo-tests") as tempdir:
            self.wrapper.home_dot_repo = tempdir
            self.wrapper.gpg_dir = os.path.join(
                self.wrapper.home_dot_repo, "gnupg"
            )
            self.assertTrue(self.wrapper.SetupGnuPG(True))
            with open(os.path.join(tempdir, "keyring-version")) as fp:
                data = fp.read()
            self.assertEqual(
                ".".join(str(x) for x in self.wrapper.KEYRING_VERSION),
                data.strip(),
            )


class VerifyRev(RepoWrapperTestCase):
    """Check verify_rev behavior."""

    def test_verify_passes(self):
        """Check when we have a valid signed tag."""
        desc_result = self.wrapper.RunResult(0, "v1.0\n", "")
        gpg_result = self.wrapper.RunResult(0, "", "")
        with mock.patch.object(
            self.wrapper, "run_git", side_effect=(desc_result, gpg_result)
        ):
            ret = self.wrapper.verify_rev(
                "/", "refs/heads/stable", "1234", True
            )
            self.assertEqual("v1.0^0", ret)

    def test_unsigned_commit(self):
        """Check we fall back to signed tag when we have an unsigned commit."""
        desc_result = self.wrapper.RunResult(0, "v1.0-10-g1234\n", "")
        gpg_result = self.wrapper.RunResult(0, "", "")
        with mock.patch.object(
            self.wrapper, "run_git", side_effect=(desc_result, gpg_result)
        ):
            ret = self.wrapper.verify_rev(
                "/", "refs/heads/stable", "1234", True
            )
            self.assertEqual("v1.0^0", ret)

    def test_verify_fails(self):
        """Check we fall back to signed tag when we have an unsigned commit."""
        desc_result = self.wrapper.RunResult(0, "v1.0-10-g1234\n", "")
        gpg_result = Exception
        with mock.patch.object(
            self.wrapper, "run_git", side_effect=(desc_result, gpg_result)
        ):
            with self.assertRaises(Exception):
                self.wrapper.verify_rev("/", "refs/heads/stable", "1234", True)


class GitCheckoutTestCase(RepoWrapperTestCase):
    """Tests that use a real/small git checkout."""

    GIT_DIR = None
    REV_LIST = None

    @classmethod
    def setUpClass(cls):
        # Create a repo to operate on, but do it once per-class.
        cls.tempdirobj = tempfile.TemporaryDirectory(prefix="repo-rev-tests")
        cls.GIT_DIR = cls.tempdirobj.name
        run_git = wrapper.Wrapper().run_git

        remote = os.path.join(cls.GIT_DIR, "remote")
        os.mkdir(remote)

        # Tests need to assume, that main is default branch at init,
        # which is not supported in config until 2.28.
        if git_command.git_require((2, 28, 0)):
            initstr = "--initial-branch=main"
        else:
            # Use template dir for init.
            templatedir = tempfile.mkdtemp(prefix=".test-template")
            with open(os.path.join(templatedir, "HEAD"), "w") as fp:
                fp.write("ref: refs/heads/main\n")
            initstr = "--template=" + templatedir

        run_git("init", initstr, cwd=remote)
        run_git("commit", "--allow-empty", "-minit", cwd=remote)
        run_git("branch", "stable", cwd=remote)
        run_git("tag", "v1.0", cwd=remote)
        run_git("commit", "--allow-empty", "-m2nd commit", cwd=remote)
        cls.REV_LIST = run_git(
            "rev-list", "HEAD", cwd=remote
        ).stdout.splitlines()

        run_git("init", cwd=cls.GIT_DIR)
        run_git(
            "fetch",
            remote,
            "+refs/heads/*:refs/remotes/origin/*",
            cwd=cls.GIT_DIR,
        )

    @classmethod
    def tearDownClass(cls):
        if not cls.tempdirobj:
            return

        cls.tempdirobj.cleanup()


class ResolveRepoRev(GitCheckoutTestCase):
    """Check resolve_repo_rev behavior."""

    def test_explicit_branch(self):
        """Check refs/heads/branch argument."""
        rrev, lrev = self.wrapper.resolve_repo_rev(
            self.GIT_DIR, "refs/heads/stable"
        )
        self.assertEqual("refs/heads/stable", rrev)
        self.assertEqual(self.REV_LIST[1], lrev)

        with self.assertRaises(self.wrapper.CloneFailure):
            self.wrapper.resolve_repo_rev(self.GIT_DIR, "refs/heads/unknown")

    def test_explicit_tag(self):
        """Check refs/tags/tag argument."""
        rrev, lrev = self.wrapper.resolve_repo_rev(
            self.GIT_DIR, "refs/tags/v1.0"
        )
        self.assertEqual("refs/tags/v1.0", rrev)
        self.assertEqual(self.REV_LIST[1], lrev)

        with self.assertRaises(self.wrapper.CloneFailure):
            self.wrapper.resolve_repo_rev(self.GIT_DIR, "refs/tags/unknown")

    def test_branch_name(self):
        """Check branch argument."""
        rrev, lrev = self.wrapper.resolve_repo_rev(self.GIT_DIR, "stable")
        self.assertEqual("refs/heads/stable", rrev)
        self.assertEqual(self.REV_LIST[1], lrev)

        rrev, lrev = self.wrapper.resolve_repo_rev(self.GIT_DIR, "main")
        self.assertEqual("refs/heads/main", rrev)
        self.assertEqual(self.REV_LIST[0], lrev)

    def test_tag_name(self):
        """Check tag argument."""
        rrev, lrev = self.wrapper.resolve_repo_rev(self.GIT_DIR, "v1.0")
        self.assertEqual("refs/tags/v1.0", rrev)
        self.assertEqual(self.REV_LIST[1], lrev)

    def test_full_commit(self):
        """Check specific commit argument."""
        commit = self.REV_LIST[0]
        rrev, lrev = self.wrapper.resolve_repo_rev(self.GIT_DIR, commit)
        self.assertEqual(commit, rrev)
        self.assertEqual(commit, lrev)

    def test_partial_commit(self):
        """Check specific (partial) commit argument."""
        commit = self.REV_LIST[0][0:20]
        rrev, lrev = self.wrapper.resolve_repo_rev(self.GIT_DIR, commit)
        self.assertEqual(self.REV_LIST[0], rrev)
        self.assertEqual(self.REV_LIST[0], lrev)

    def test_unknown(self):
        """Check unknown ref/commit argument."""
        with self.assertRaises(self.wrapper.CloneFailure):
            self.wrapper.resolve_repo_rev(self.GIT_DIR, "boooooooya")


class CheckRepoVerify(RepoWrapperTestCase):
    """Check check_repo_verify behavior."""

    def test_no_verify(self):
        """Always fail with --no-repo-verify."""
        self.assertFalse(self.wrapper.check_repo_verify(False))

    def test_gpg_initialized(self):
        """Should pass if gpg is setup already."""
        with mock.patch.object(
            self.wrapper, "NeedSetupGnuPG", return_value=False
        ):
            self.assertTrue(self.wrapper.check_repo_verify(True))

    def test_need_gpg_setup(self):
        """Should pass/fail based on gpg setup."""
        with mock.patch.object(
            self.wrapper, "NeedSetupGnuPG", return_value=True
        ):
            with mock.patch.object(self.wrapper, "SetupGnuPG") as m:
                m.return_value = True
                self.assertTrue(self.wrapper.check_repo_verify(True))

                m.return_value = False
                self.assertFalse(self.wrapper.check_repo_verify(True))


class CheckRepoRev(GitCheckoutTestCase):
    """Check check_repo_rev behavior."""

    def test_verify_works(self):
        """Should pass when verification passes."""
        with mock.patch.object(
            self.wrapper, "check_repo_verify", return_value=True
        ):
            with mock.patch.object(
                self.wrapper, "verify_rev", return_value="12345"
            ):
                rrev, lrev = self.wrapper.check_repo_rev(self.GIT_DIR, "stable")
        self.assertEqual("refs/heads/stable", rrev)
        self.assertEqual("12345", lrev)

    def test_verify_fails(self):
        """Should fail when verification fails."""
        with mock.patch.object(
            self.wrapper, "check_repo_verify", return_value=True
        ):
            with mock.patch.object(
                self.wrapper, "verify_rev", side_effect=Exception
            ):
                with self.assertRaises(Exception):
                    self.wrapper.check_repo_rev(self.GIT_DIR, "stable")

    def test_verify_ignore(self):
        """Should pass when verification is disabled."""
        with mock.patch.object(
            self.wrapper, "verify_rev", side_effect=Exception
        ):
            rrev, lrev = self.wrapper.check_repo_rev(
                self.GIT_DIR, "stable", repo_verify=False
            )
        self.assertEqual("refs/heads/stable", rrev)
        self.assertEqual(self.REV_LIST[1], lrev)
